package be.iminds.ilabt.jfed.testing.tests.highlevel;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.rspec.AdvertisementBasedRspecGeneratorConfig;
import be.iminds.ilabt.jfed.testing.base.ApiTestConfig;
import be.iminds.ilabt.jfed.testing.base.ApiTestConfigProperty;
import be.iminds.ilabt.jfed.testing.shared.NodeLoginTestStepConfig;
import com.fasterxml.jackson.annotation.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.jetbrains.annotations.Contract;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.*;
import java.util.function.Supplier;

@JsonInclude(JsonInclude.Include.NON_DEFAULT)
@JsonIgnoreProperties(value={"@context"})
public class GuiLogicTestConfig implements ApiTestConfig {
    private static final Logger LOG = LoggerFactory.getLogger(GuiLogicTestConfig.class);

    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    @Nonnull private final be.iminds.ilabt.jfed.testing.shared.Proxy amConnectionProxy;
    @Nonnull private final String amVersion;
    @Nonnull private final SSHKeys sshKeys;
    @Nonnull private final List<Resource> resources;
    @Nonnull private final List<AnsibleTest> ansibleTests;
    @Nonnull private final NodeLoginTest nodeLoginTest;

    public GuiLogicTestConfig() {
        this.amConnectionProxy = new be.iminds.ilabt.jfed.testing.shared.Proxy();
        this.amVersion = "auto";
        this.sshKeys = new SSHKeys();
        this.resources = Collections.singletonList(new Resource());
        this.ansibleTests = Collections.singletonList(new AnsibleTest());
        this.nodeLoginTest = new NodeLoginTest();
    }

    public GuiLogicTestConfig(@Nullable /*@JsonProperty("amConnectionProxy")*/ be.iminds.ilabt.jfed.testing.shared.Proxy amConnectionProxy,
                              @Nullable /*@JsonProperty("amVersion")        */ String amVersion,
                              @Nullable /*@JsonProperty("sshKeys")          */ SSHKeys sshKeys,
                              @Nullable /*@JsonProperty("resources")        */ List<Resource> resources,
                              @Nullable /*@JsonProperty("ansibleTests")     */ List<AnsibleTest> ansibleTests,
                              @Nullable /*@JsonProperty("nodeLoginTest")    */ NodeLoginTest nodeLoginTest) {
        this.amConnectionProxy = amConnectionProxy == null ? new GuiLogicTestConfig().getAmConnectionProxy() : amConnectionProxy;
        this.amVersion = amVersion == null ? new GuiLogicTestConfig().getAmVersion() : amVersion;
        this.sshKeys = sshKeys == null ? new GuiLogicTestConfig().getSshKeys() : sshKeys;
        this.resources = resources == null ? new GuiLogicTestConfig().getResources() : resources;
        this.ansibleTests = ansibleTests == null ? new GuiLogicTestConfig().getAnsibleTests() : ansibleTests;
        this.nodeLoginTest = nodeLoginTest == null ? new GuiLogicTestConfig().getNodeLoginTest() : nodeLoginTest;
    }

    public GuiLogicTestConfig(@Nullable /*@JsonProperty("amConnectionProxy")*/ be.iminds.ilabt.jfed.testing.shared.Proxy amConnectionProxy,
                              @Nullable /*@JsonProperty("amVersion")        */ String amVersion,
                              @Nullable /*@JsonProperty("sshKeys")          */ SSHKeys sshKeys,
                              @Nullable /*@JsonProperty("resources")        */ List<Resource> resources,
                              @Nullable /*@JsonProperty("ansibleTests")     */ AnsibleTest ansibleTest,
                              @Nullable /*@JsonProperty("nodeLoginTest")    */ NodeLoginTest nodeLoginTest) {
        this(amConnectionProxy, amVersion, sshKeys, resources,
                ansibleTest == null ?
                        new GuiLogicTestConfig().getAnsibleTests() :
                        Collections.singletonList(ansibleTest),
                nodeLoginTest);}

    @JsonCreator
    public static GuiLogicTestConfig createFromJsonObject(@Nonnull Map<String,Object> props) {
        be.iminds.ilabt.jfed.testing.shared.Proxy proxy = MAPPER.convertValue(props.get("amConnectionProxy"), be.iminds.ilabt.jfed.testing.shared.Proxy.class);
        String amVersion = (String) props.get("amVersion");
        SSHKeys sshKeys = MAPPER.convertValue(props.get("sshKeys"), SSHKeys.class);
        List<Resource> resources = MAPPER.convertValue(props.get("resources"), new TypeReference<List<Resource>>() { });;
        List<AnsibleTest> ansibleTests;
        NodeLoginTest nodeLoginTest = MAPPER.convertValue(props.get("nodeLoginTest"), NodeLoginTest.class);

        if (props.containsKey("ansibleTest") || props.containsKey("ansibleTests")) {
            String key = props.containsKey("ansibleTests") ? "ansibleTests" : "ansibleTest";
            Object ansibleTestsObj = props.get(key);
            if (ansibleTestsObj instanceof List) {
                ansibleTests = MAPPER.convertValue(ansibleTestsObj, new TypeReference<List<AnsibleTest>>() { });
            } else {
                ansibleTests = Collections.singletonList(
                        MAPPER.convertValue(ansibleTestsObj, AnsibleTest.class)
                );
            }
        } else {
            ansibleTests = null;
        }

        return new GuiLogicTestConfig(proxy, amVersion, sshKeys, resources, ansibleTests, nodeLoginTest);
    }

    @Override
    @JsonIgnore
    public boolean isValid() {
        boolean res = getNodeLoginTest().isValid();
        res = res && getAmConnectionProxy().isValid();
        res = res && getSshKeys().isValid();
        for (Resource resource : getResources()) {
            res = res && resource.isValid();
        }
        for (AnsibleTest at : getAnsibleTests()) {
            res = res && at.isValid();
        }
        boolean amVersionValid = getAmVersion().equals("2") || getAmVersion().equals("3") || getAmVersion().equals("auto");
        if (!amVersionValid)
            throw new IllegalArgumentException("Config Error: Unsupported amVersion: \""+getAmVersion()+"\"");
//        res = res && amVersionValid;
        return res;
    }

    @JsonProperty
    @Nonnull
    public be.iminds.ilabt.jfed.testing.shared.Proxy getAmConnectionProxy() {
        return amConnectionProxy;
    }

    @JsonProperty
    @Nonnull
    public String getAmVersion() {
        return amVersion;
    }

    @JsonProperty
    @Nonnull
    public SSHKeys getSshKeys() {
        return sshKeys;
    }

    @JsonProperty
    @Nonnull
    public List<Resource> getResources() {
        return resources;
    }

    @JsonProperty
    @Nonnull
    public List<AnsibleTest> getAnsibleTests() {
        return ansibleTests;
    }

    @JsonProperty
    @Nonnull
    public NodeLoginTest getNodeLoginTest() {
        return nodeLoginTest;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof GuiLogicTestConfig)) return false;

        GuiLogicTestConfig that = (GuiLogicTestConfig) o;

        if (!amConnectionProxy.equals(that.amConnectionProxy)) return false;
        if (!amVersion.equals(that.amVersion)) return false;
        if (!sshKeys.equals(that.sshKeys)) return false;
        if (!resources.equals(that.resources)) return false;
        if (!ansibleTests.equals(that.ansibleTests)) return false;
        return nodeLoginTest.equals(that.nodeLoginTest);
    }

    @Override
    public int hashCode() {
        int result = amConnectionProxy.hashCode();
        result = 31 * result + amVersion.hashCode();
        result = 31 * result + sshKeys.hashCode();
        result = 31 * result + resources.hashCode();
        result = 31 * result + ansibleTests.hashCode();
        result = 31 * result + nodeLoginTest.hashCode();
        return result;
    }

    @Nonnull
    @Override
    @JsonIgnore
    public String stringValue() {
        try {
            return MAPPER.writeValueAsString(this);
        } catch (JsonProcessingException e) {
            throw new RuntimeException("Failed to covert "+getClass().getName()+" to String", e);
        }
    }

    @Override
    public String toString() {
        return stringValue();
    }

    //What sort of data is captured/extracted?
    public enum ExtractDataType {
        TEXT, CSV, JSON
    }
    //What is the captured/extracted data about?
    public enum ExtractScope {
        TESTINSTANCE, SERVER, NODE
    }

    public static class ExtractSpec {
        //What to extract
        @Nonnull private final String delim;
        @Nullable private final String startDelim;
        @Nullable private final String endDelim;
        @Nullable private final String regex;    //a regex with a group, that is used to extract. Used instead of delim if present.
        @Nonnull private final ExtractDataType type;

        //Where to store it
        @Nonnull private final String name;
        @Nonnull private final ExtractScope scope;

        public ExtractSpec() {
            //startDelim and endDelim take priority over delim, if present
            //regex takes priority over all if present!
            this.delim = "$$$";
            this.startDelim = null;
            this.endDelim = null;
            this.regex = null;
            this.type = ExtractDataType.TEXT;
            this.name = "data";
            this.scope = ExtractScope.TESTINSTANCE;
        }

        @JsonCreator
        public ExtractSpec(@Nullable @JsonProperty("delim") String delim,
                           @Nullable @JsonProperty("startDelim") String startDelim,
                           @Nullable @JsonProperty("endDelim") String endDelim,
                           @Nullable @JsonProperty("regex") String regex,
                           @Nullable @JsonProperty("type") ExtractDataType type,
                           @Nullable @JsonProperty("name") String name,
                           @Nullable @JsonProperty("scope") ExtractScope scope) {
            this.delim = delim != null ? delim : new ExtractSpec().getDelim();
            this.startDelim = startDelim == null || startDelim.trim().isEmpty() ? null : startDelim;
            this.endDelim = endDelim == null || endDelim.trim().isEmpty() ? null : endDelim;
            this.regex = regex == null || regex.trim().isEmpty() ? null : regex;
            this.type = type != null ? type : new ExtractSpec().getType();
            this.name = name != null ? name : new ExtractSpec().getName();
            this.scope = scope != null ? scope : new ExtractSpec().getScope();
        }

        @JsonProperty
        @Nonnull
        public String getDelim() {
            return delim;
        }

        @JsonProperty
        @Nullable
        public String getStartDelim() {
            return startDelim;
        }

        @JsonProperty
        @Nullable
        public String getEndDelim() {
            return endDelim;
        }

        @JsonProperty
        @Nullable
        public String getRegex() {
            return regex;
        }

        @JsonIgnore
        @Nonnull
        public String getStartDelimOrFallback() {
            return startDelim == null ? delim : startDelim;
        }

        @JsonIgnore
        @Nonnull
        public String getEndDelimOrFallback() {
            return endDelim == null ? delim : endDelim;
        }

        @JsonProperty
        @Nonnull
        public ExtractDataType getType() {
            return type;
        }

        @JsonProperty
        @Nonnull
        public String getName() {
            return name;
        }

        @JsonProperty
        @Nonnull
        public ExtractScope getScope() {
            return scope;
        }

        @JsonIgnore
        public boolean isValid() {
            //delim is fallback, so may never be null
            if (delim.trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.extract.delim may not be empty");
            }
            if (name.trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.extract.name may not be empty");
            }
            return true;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context", "playbookExe"})
    public static class AnsibleTest {
        @Nonnull private final Boolean enabled;
        @Nonnull private final Boolean debug;
        @Nullable private final String playbookUrl; //may be yml file, or archive (zip, tar, ...)
        @Nullable private final String playbookFileNameInArchive; //if playbook URL points to an archive, this is the name of the ansible playbook in it
        @Nullable private final String playbookContent;
        @Nullable private final String failureRegex;
        @Nullable private final String warningRegex;
        @Nullable private final String successRegex;
        @Nonnull private final be.iminds.ilabt.jfed.testing.shared.Proxy proxy;
        @Nonnull private final List<ExtractSpec> extract;
        private final int timeoutInSec; //<= 0 means no timeout

        public AnsibleTest() {
            this.enabled = false;
            this.proxy = new be.iminds.ilabt.jfed.testing.shared.Proxy();
            this.debug = false;
            this.playbookUrl = null;
            this.playbookFileNameInArchive = null;
            this.playbookContent = null;
            this.failureRegex = "(failed=[^0]|unreachable=[^0])"; //"failed=[^0]";
            this.warningRegex = null;//"unreachable=[^0]";
            this.successRegex = "ok=[^0]";
            this.extract = Collections.emptyList();
            this.timeoutInSec = 20*60;
        }

        @JsonCreator
        public AnsibleTest(@Nullable @JsonProperty("enabled") Boolean enabled,
                           @Nullable @JsonProperty("proxy") be.iminds.ilabt.jfed.testing.shared.Proxy proxy,
                           @Nullable @JsonProperty("debug") Boolean debug,
                           @Nullable @JsonProperty("playbookUrl") String playbookUrl,
                           @Nullable @JsonProperty("playbookFileNameInArchive") String playbookFileNameInArchive,
                           @Nullable @JsonProperty("playbookContent") String playbookContent,
                           @Nullable @JsonProperty("failureRegex") String failureRegex,
                           @Nullable @JsonProperty("warningRegex") String warningRegex,
                           @Nullable @JsonProperty("successRegex") String successRegex,
                           @Nullable @JsonProperty("extract") List<ExtractSpec> extract,
                           @Nullable @JsonProperty("timeoutInSec") Integer timeoutInSec) {
            this.enabled = enabled == null ? new AnsibleTest().isEnabled() : enabled;
            this.proxy = proxy == null ? new AnsibleTest().getProxy() : proxy;
            this.debug = debug == null ? new AnsibleTest().getDebug() : debug;
            this.playbookUrl = trimAndToNullIfEmpty(playbookUrl);
            this.playbookFileNameInArchive = trimAndToNullIfEmpty(playbookFileNameInArchive);
            this.playbookContent = toNullIfEmpty(playbookContent);
            this.failureRegex = toNullableDefaultIfEmpty(failureRegex, () -> new AnsibleTest().getFailureRegex());
            this.warningRegex = toNullableDefaultIfEmpty(warningRegex, () -> new AnsibleTest().getWarningRegex());
            this.successRegex = toNullableDefaultIfEmpty(successRegex, () -> new AnsibleTest().getSuccessRegex());
            this.extract = extract != null ?
                    Collections.unmodifiableList(new ArrayList<>(extract)) :
                    new AnsibleTest().getExtract();
            this.timeoutInSec = timeoutInSec == null ? 20*60 : timeoutInSec;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!isEnabled())
                return true;

            if (getPlaybookContent() == null && getPlaybookUrl() == null) {
                throw new IllegalArgumentException("Config Error: You need to specify either ansible.playbookContent or ansible.playbookUrl");
            }
            if (getPlaybookContent() != null && getPlaybookUrl() != null) {
                throw new IllegalArgumentException("Config Error: You need to specify either ansible.playbookContent or ansible.playbookUrl, but not both");
            }
            if (getPlaybookContent() != null && getPlaybookContent().trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.playbookContent may not be empty when specified");
            }
            if (getPlaybookUrl() != null) {
                if (getPlaybookUrl().trim().isEmpty()) {
                    throw new IllegalArgumentException("Config Error: ansible.playbookUrl may not be empty when specified");
                }
                if (!(getPlaybookUrl().trim().startsWith("http://") || getPlaybookUrl().trim().startsWith("https://"))) {
                    if (getPlaybookUrl().trim().startsWith("git ") || getPlaybookUrl().trim().startsWith("github ")) {
                        //ok, special case: git/github with subdir and branch
                        //subdir may be / for no subdir

                        //syntax: git[hub] <https://<githubTLD>/<org>/<repo>.git> [<subdir> [<branch>]]
                        if (playbookFileNameInArchive == null) {
                            throw new IllegalArgumentException("Config Error: ansible.playbookUrl refers to git (\"" + getPlaybookUrl() + "\"), but playbookFileNameInArchive is not set");
                        }
                    } else {
                        throw new IllegalArgumentException("Config Error: ansible.playbookUrl does not appear to be a URL: \"" + getPlaybookUrl() + "\"");
                    }
                } else {
                    if (isProbablyArchiveUrl(getPlaybookUrl().trim()) && playbookFileNameInArchive == null) {
                        throw new IllegalArgumentException("Config Error: ansible.playbookUrl appear to be an archive(\"" + getPlaybookUrl() + "\"), but playbookFileNameInArchive is not set");
                    }
                }
            }
            if (getPlaybookUrl() == null && playbookFileNameInArchive != null) {
                throw new IllegalArgumentException("Config Error: ansible.playbookFileNameInArchive only makes sense if ansible.playbookUrl is used (and points to an archive)");
            }
            if (getFailureRegex() != null && getFailureRegex().trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.failureRegex may not be empty");
            }
            if (getWarningRegex() != null && getWarningRegex().trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.warningRegex may not be empty");
            }
            if (getSuccessRegex() != null && getSuccessRegex().trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.successRegex may not be empty");
            }

            for (ExtractSpec extractSpec : getExtract()) {
                boolean extractValid = extractSpec.isValid();
                if (!extractValid)
                    return false;
            }

            return proxy.isValid();
        }

        private static boolean isProbablyArchiveUrl(String url) {
            if (url.endsWith(".zip"))
                return true;
            if (url.endsWith(".tar"))
                return true;
            if (url.endsWith(".tar.gz"))
                return true;
            if (url.endsWith(".tar.bz2"))
                return true;
            if (url.endsWith(".rar"))
                return true;
            return false;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Timeout of the test in seconds. <= 0 means to timeout. Default is 20 minutes.")
        public int getTimeoutInSec() {
            return timeoutInSec;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="enables the runAnsible() step of the test")
        public boolean isEnabled() {
            return enabled;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="(SSH) proxy to use for the ansible tests")
        @Nonnull
        public be.iminds.ilabt.jfed.testing.shared.Proxy getProxy() {
            return proxy;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="sets the debug flag when calling ansible")
        public boolean getDebug() {
            return debug;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="URL of the PlayBook to execute (either playbookUrl or playbookUrl must be provided, and not both)")
        @Nullable
        public String getPlaybookUrl() {
            return playbookUrl;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Name of the PlayBook inside the archive pointed to by playbookUrl")
        @Nullable
        public String getPlaybookFileNameInArchive() {
            return playbookFileNameInArchive;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Content of the PlayBook to execute (either playbookUrl or playbookUrl must be provided, and not both)")
        @Nullable
        public String getPlaybookContent() {
            return playbookContent;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="When this regex matches a line in the ansible output, the test state is considered failed.")
        @Nullable
        public String getFailureRegex() {
            return failureRegex;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="When this regex matches a line in the ansible output, the test state is considered warning. (Unless it is failed.)")
        @Nullable
        public String getWarningRegex() {
            return warningRegex;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="When this regex matches a line in the ansible output, the test state is considered succes. (Unless it is failed or warning.)")
        @Nullable
        public String getSuccessRegex() {
            return successRegex;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Data to extract from ansible result")
        @Nonnull
        public List<ExtractSpec> getExtract() {
            return extract;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof AnsibleTest)) return false;

            AnsibleTest that = (AnsibleTest) o;

            if (!enabled.equals(that.enabled)) return false;
            if (!debug.equals(that.debug)) return false;
            if (playbookUrl != null ? !playbookUrl.equals(that.playbookUrl) : that.playbookUrl != null) return false;
            if (playbookFileNameInArchive != null ? !playbookFileNameInArchive.equals(that.playbookFileNameInArchive) : that.playbookFileNameInArchive != null)
                return false;
            if (playbookContent != null ? !playbookContent.equals(that.playbookContent) : that.playbookContent != null)
                return false;
            if (failureRegex != null ? !failureRegex.equals(that.failureRegex) : that.failureRegex != null)
                return false;
            if (warningRegex != null ? !warningRegex.equals(that.warningRegex) : that.warningRegex != null)
                return false;
            if (successRegex != null ? !successRegex.equals(that.successRegex) : that.successRegex != null)
                return false;
            if (!proxy.equals(that.proxy)) return false;
            return extract.equals(that.extract);
        }

        @Override
        public int hashCode() {
            int result = enabled.hashCode();
            result = 31 * result + debug.hashCode();
            result = 31 * result + (playbookUrl != null ? playbookUrl.hashCode() : 0);
            result = 31 * result + (playbookFileNameInArchive != null ? playbookFileNameInArchive.hashCode() : 0);
            result = 31 * result + (playbookContent != null ? playbookContent.hashCode() : 0);
            result = 31 * result + (failureRegex != null ? failureRegex.hashCode() : 0);
            result = 31 * result + (warningRegex != null ? warningRegex.hashCode() : 0);
            result = 31 * result + (successRegex != null ? successRegex.hashCode() : 0);
            result = 31 * result + proxy.hashCode();
            result = 31 * result + extract.hashCode();
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class NodeLoginTest {
        private final boolean enabled;
        private final long deadlineSeconds;
        @Nullable private final String preferredUser;
        @Nullable private final Integer forceIpVersion;
        private final boolean useExternalSsh;
        @Nonnull private final be.iminds.ilabt.jfed.testing.shared.Proxy proxy;

        public NodeLoginTest() {
            this.enabled = true;
            this.deadlineSeconds = 20 * 60;
            this.preferredUser = null;
            this.forceIpVersion = null;
            this.useExternalSsh = false;
            this.proxy = new be.iminds.ilabt.jfed.testing.shared.Proxy();
        }

        @JsonCreator
        public NodeLoginTest(@Nullable @JsonProperty("enabled") Boolean enabled,
                             @Nullable @JsonProperty("deadlineSeconds") Long deadlineSeconds,
                             @Nullable @JsonProperty("preferredUser") String preferredUser,
                             @Nullable @JsonProperty("forceIpVersion") Integer forceIpVersion,
                             @Nullable @JsonProperty("useExternalSsh") Boolean useExternalSsh,
                             @Nullable @JsonProperty("proxy") be.iminds.ilabt.jfed.testing.shared.Proxy proxy) {
            this.enabled = enabled == null ? new NodeLoginTest().isEnabled() : enabled;
            this.deadlineSeconds = deadlineSeconds == null ? new NodeLoginTest().getDeadlineSeconds() : deadlineSeconds;
            this.preferredUser = preferredUser;
            this.forceIpVersion = forceIpVersion;
            this.useExternalSsh = useExternalSsh == null ? new NodeLoginTest().getUseExternalSsh() : useExternalSsh;
            this.proxy = proxy == null ? new NodeLoginTest().getProxy() : proxy;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!isEnabled())
                return true;

            if (deadlineSeconds <= 0)
                throw new IllegalArgumentException("Config Error: " +
                        "nodelogin.deadlineSeconds <= 0 makes no sense: "+deadlineSeconds);

            boolean ipVersionValid = getForceIpVersion() == null ||
                    getForceIpVersion().equals(4) ||
                    getForceIpVersion().equals(6);
            if (!ipVersionValid)
                throw new IllegalArgumentException("Config Error: " +
                        "Unsupported forceIpVersion: \""+getForceIpVersion()+"\"");

            return proxy.isValid();
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @JsonProperty
        public long getDeadlineSeconds() {
            return deadlineSeconds;
        }

        @JsonProperty
        @Nullable
        public String getPreferredUser() {
            return preferredUser;
        }

        @JsonProperty
        @Nullable
        public Integer getForceIpVersion() {
            return forceIpVersion;
        }

        @JsonProperty
        public boolean getUseExternalSsh() {
            return useExternalSsh;
        }

        @JsonProperty
        @Nonnull
        public be.iminds.ilabt.jfed.testing.shared.Proxy getProxy() {
            return proxy;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof NodeLoginTest)) return false;

            NodeLoginTest that = (NodeLoginTest) o;

            if (enabled != that.enabled) return false;
            if (deadlineSeconds != that.deadlineSeconds) return false;
            if (useExternalSsh != that.useExternalSsh) return false;
            if (preferredUser != null ? !preferredUser.equals(that.preferredUser) : that.preferredUser != null)
                return false;
            if (forceIpVersion != null ? !forceIpVersion.equals(that.forceIpVersion) : that.forceIpVersion != null)
                return false;
            return proxy.equals(that.proxy);
        }

        @Override
        public int hashCode() {
            int result = (enabled ? 1 : 0);
            result = 31 * result + (int) (deadlineSeconds ^ (deadlineSeconds >>> 32));
            result = 31 * result + (preferredUser != null ? preferredUser.hashCode() : 0);
            result = 31 * result + (forceIpVersion != null ? forceIpVersion.hashCode() : 0);
            result = 31 * result + (useExternalSsh ? 1 : 0);
            result = 31 * result + proxy.hashCode();
            return result;
        }
    }

    public enum SSHKeySource { TEST_USER, RANDOM };

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class SSHKeys {
        @Nonnull private final SSHKeySource source;

        //maybe future, but not sure how to do this securily from fedmon:
        // , fixedContent, fixedFile
//        @Nonnull private final String publicKeyContent;
//        @Nonnull private final String privateKeyContent;
//        @Nonnull private final String publicKeyFile;
//        @Nonnull private final String privateKeyFile;
//        @Nonnull private final String privateKeyPassword;


        public SSHKeys() {
            this.source = SSHKeySource.RANDOM;
        }

        @JsonCreator
        public SSHKeys(@Nullable @JsonProperty("source") SSHKeySource source) {
            this.source = source == null ? new SSHKeys().getSource() : source;
        }

        @JsonIgnore
        public boolean isValid() {
            return true;
        }

        @JsonProperty
        @Nonnull
        public SSHKeySource getSource() {
            return source;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof SSHKeys)) return false;

            SSHKeys sshKeys = (SSHKeys) o;

            return source == sshKeys.source;
        }

        @Override
        public int hashCode() {
            return source.hashCode();
        }
    }

    public enum RequestRSpecSource { PROVIDE_URL, PROVIDE_FILE, PROVIDE_CONTENT,
        GENERATE_USING_JFED, GENERATE_SIMPLE, GENERATE_USING_ADVERTISEMENT,
        ALREADY_PROVISIONED;

        public boolean isProvide() {
            return this == PROVIDE_URL ||
                    this == PROVIDE_FILE ||
                    this == PROVIDE_CONTENT;
        }
    }

    @JsonTypeInfo(
            use = JsonTypeInfo.Id.NAME,
            include = JsonTypeInfo.As.PROPERTY,
            property = "source",
            defaultImpl = GenerateUsingJFedRequestRSpec.class,
            visible = true)
    @JsonSubTypes({
            @JsonSubTypes.Type(value = GenerateSimpleRequestRSpec.class, name = "GENERATE_SIMPLE"),
            @JsonSubTypes.Type(value = ProvidedUrlRequestRSpec.class, name = "PROVIDE_URL"),
            @JsonSubTypes.Type(value = ProvidedFileRequestRSpec.class, name = "PROVIDE_FILE"),
            @JsonSubTypes.Type(value = ProvidedContentRequestRSpec.class, name = "PROVIDE_CONTENT"),
            @JsonSubTypes.Type(value = GenerateUsingJFedRequestRSpec.class, name = "GENERATE_USING_JFED"),
            @JsonSubTypes.Type(value = GenerateUsingAdvertisementRequestRSpec.class, name = "GENERATE_USING_ADVERTISEMENT"),
            @JsonSubTypes.Type(value = AlreadyProvisionedRequestRSpec.class, name = "ALREADY_PROVISIONED")
    })
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static abstract class RequestRSpec {
        @Nonnull protected final RequestRSpecSource source;

        //used for all methods
        private final boolean autoAssignSpecificNodesUsingListResources;
        private final boolean useNitosLease;

        @JsonCreator
        public RequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                            @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                            @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            assert source != null;
            this.source = source;
            this.autoAssignSpecificNodesUsingListResources =
                    autoAssignSpecificNodesUsingListResources == null ?
                            getDefault().getAutoAssignSpecificNodesUsingListResources() :
                            autoAssignSpecificNodesUsingListResources;
            this.useNitosLease = useNitosLease == null ? getDefault().getUseNitosLease() : useNitosLease;
        }

        public static RequestRSpec getDefault() {
            return new GenerateUsingJFedRequestRSpec();
        }


        @JsonCreator
        public static RequestRSpec createRequestRSpecFromJsonObject(@Nonnull Map<String,Object> props) {
            //TODO: Check if types are correct etc
            RequestRSpecSource source =
                    props.get("source") == null ?
                            null :
                            RequestRSpecSource.valueOf((String) props.get("source"));
            Boolean autoAssignSpecificNodesUsingListResources = (Boolean) props.get("autoAssignSpecificNodesUsingListResources");
            Boolean useNitosLease = (Boolean) props.get("useNitosLease");

            LOG.debug("createRequestRSpecFromJsonObject with source="+source);

            if (source == null) {
                LOG.debug("createRequestRSpecFromJsonObject with source=null -> reverting all to defaults");
                return getDefault();
            }

            switch (source) {
                case ALREADY_PROVISIONED: {
                    return new AlreadyProvisionedRequestRSpec(
                            source,
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease);
                }
                case PROVIDE_FILE: {
                    return new ProvidedFileRequestRSpec(source,
                            (String) props.get("providedContentSource"),
                            (Boolean) props.get("bindTestedServerToUnboundNodes"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease
                    );
                }
                case PROVIDE_URL: {
                    return new ProvidedUrlRequestRSpec(source,
                            (String) props.get("providedContentSource"),
                            (Boolean) props.get("bindTestedServerToUnboundNodes"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease
                    );
                }
                case PROVIDE_CONTENT: {
                    return new ProvidedContentRequestRSpec(source,
                            (String) props.get("providedContentSource"),
                            (Boolean) props.get("bindTestedServerToUnboundNodes"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease
                    );
                }
                case GENERATE_USING_ADVERTISEMENT: {
                    AdvertisementBasedRspecGeneratorConfig advertisementBasedRspecGeneratorConfig =
                            MAPPER.convertValue(
                                    props.get("config"),
                                    AdvertisementBasedRspecGeneratorConfig.class);
                    return new GenerateUsingAdvertisementRequestRSpec(source,
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease,
                            advertisementBasedRspecGeneratorConfig);
                }
                case GENERATE_SIMPLE: {
                    return new GenerateSimpleRequestRSpec(source,
                            (Integer) props.get("nodeCount"),
                            (String) props.get("nodeSliverType"),
                            (Boolean) props.get("nodeExclusive"),
                            (List<String>) props.get("nodeComponentUrn"),
                            (List<String>) props.get("nodeAnsibleGroup"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease);
                }
                case GENERATE_USING_JFED: {
                    return new GenerateUsingJFedRequestRSpec(source,
                            (Integer) props.get("nodeCount"),
                            (String) props.get("jFedResourceClass"),
                            (List<String>) props.get("nodeAnsibleGroup"),
                            autoAssignSpecificNodesUsingListResources,
                            useNitosLease);
                }
                default:
                    throw new IllegalArgumentException("Config Error: unsupported rspec.source == "+source);
            }
        }

        @JsonIgnore
        public abstract boolean isValid();

        @JsonProperty
        @Nonnull
        public RequestRSpecSource getSource() {
            return source;
        }

        @JsonProperty
        public boolean getAutoAssignSpecificNodesUsingListResources() {
            return autoAssignSpecificNodesUsingListResources;
        }

        @JsonProperty
        public boolean getUseNitosLease() {
            return useNitosLease;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof RequestRSpec)) return false;

            RequestRSpec that = (RequestRSpec) o;

            if (autoAssignSpecificNodesUsingListResources != that.autoAssignSpecificNodesUsingListResources)
                return false;
            if (useNitosLease != that.useNitosLease) return false;
            if (source != that.source) return false;
            return true;
        }

        @Override
        public int hashCode() {
            int result = source.hashCode();
            result = 31 * result + (autoAssignSpecificNodesUsingListResources ? 1 : 0);
            result = 31 * result + (useNitosLease ? 1 : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class GenerateUsingJFedRequestRSpec extends RequestRSpec {
        private final int nodeCount;
        @Nullable private final String jFedResourceClass; //if null, a class will be picked automatically, using the order in the jFed GUI (top being highest priority)
        @Nonnull private final List<String> nodeAnsibleGroup;

        public GenerateUsingJFedRequestRSpec() {
            super(RequestRSpecSource.GENERATE_USING_JFED, false, false);
            this.nodeCount = 1;
            this.jFedResourceClass = null;
            this.nodeAnsibleGroup = Collections.emptyList();
        }

        @JsonCreator
        public GenerateUsingJFedRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                             @Nullable @JsonProperty("nodeCount") Integer nodeCount,
                                             @Nullable @JsonProperty("jFedResourceClass") String jFedResourceClass,
                                             @Nullable @JsonProperty("nodeAnsibleGroup") List<String> nodeAnsibleGroup,
                                             @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                             @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(RequestRSpecSource.GENERATE_USING_JFED, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.GENERATE_USING_JFED;
            assert this.getSource() == RequestRSpecSource.GENERATE_USING_JFED;
            this.nodeCount = nodeCount == null ? new GenerateUsingJFedRequestRSpec().getNodeCount() : nodeCount;
            this.jFedResourceClass = jFedResourceClass == null ? new GenerateUsingJFedRequestRSpec().getjFedResourceClass() : jFedResourceClass;
            this.nodeAnsibleGroup = nodeAnsibleGroup == null ? new GenerateSimpleRequestRSpec().getNodeAnsibleGroup() : nodeAnsibleGroup;
        }

        @JsonIgnore
        public boolean isValid() {
            if (nodeCount <= 0)
                throw new IllegalArgumentException("Config Error: rspec.nodeCount == "+nodeCount+" makes no sense if rspec.source == "+source);
            return true;
        }

        @JsonProperty
        public int getNodeCount() {
            return nodeCount;
        }

        @JsonProperty
        @Nullable
        public String getjFedResourceClass() {
            return jFedResourceClass;
        }

        @JsonProperty
        @Nonnull
        public List<String> getNodeAnsibleGroup() {
            return nodeAnsibleGroup;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof GenerateUsingJFedRequestRSpec)) return false;
            if (!super.equals(o)) return false;

            GenerateUsingJFedRequestRSpec that = (GenerateUsingJFedRequestRSpec) o;

            if (nodeCount != that.nodeCount) return false;
            if (!nodeAnsibleGroup.equals(that.nodeAnsibleGroup)) return false;
            return jFedResourceClass != null ? jFedResourceClass.equals(that.jFedResourceClass) : that.jFedResourceClass == null;
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            result = 31 * result + nodeCount;
            result = 31 * result + (jFedResourceClass != null ? jFedResourceClass.hashCode() : 0);
            result = 31 * result + nodeAnsibleGroup.hashCode();
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class AlreadyProvisionedRequestRSpec extends RequestRSpec {
        public AlreadyProvisionedRequestRSpec() {
            super(RequestRSpecSource.ALREADY_PROVISIONED,
                    false,
                    false);
        }

        @JsonCreator
        public AlreadyProvisionedRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                              @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                              @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.ALREADY_PROVISIONED;
            assert this.getSource() == RequestRSpecSource.ALREADY_PROVISIONED;
        }

        @JsonIgnore
        public boolean isValid() {
            return true;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ProvidedContentRequestRSpec extends ProvidedRequestRSpec {
        public ProvidedContentRequestRSpec() {
            super(RequestRSpecSource.PROVIDE_CONTENT,
                    null,
                    false,
                    false,
                    false);
        }

        @JsonCreator
        public ProvidedContentRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                           @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                                           @Nullable @JsonProperty("bindTestedServerToUnboundNodes") Boolean bindTestedServerToUnboundNodes,
                                           @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                           @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, providedContentSource, bindTestedServerToUnboundNodes,
                    autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.PROVIDE_CONTENT;
            assert this.getSource() == RequestRSpecSource.PROVIDE_CONTENT;
        }
    }
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ProvidedUrlRequestRSpec extends ProvidedRequestRSpec {
        public ProvidedUrlRequestRSpec() {
            super(RequestRSpecSource.PROVIDE_URL,
                    null,
                    false,
                    false,
                    false);
        }

        @JsonCreator
        public ProvidedUrlRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                       @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                                       @Nullable @JsonProperty("bindTestedServerToUnboundNodes") Boolean bindTestedServerToUnboundNodes,
                                       @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                       @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, providedContentSource, bindTestedServerToUnboundNodes,
                    autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.PROVIDE_URL;
            assert this.getSource() == RequestRSpecSource.PROVIDE_URL;
        }
    }
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ProvidedFileRequestRSpec extends ProvidedRequestRSpec {
        public ProvidedFileRequestRSpec() {
            super(RequestRSpecSource.PROVIDE_FILE,
                    null,
                    false,
                    false,
                    false);
        }

        @JsonCreator
        public ProvidedFileRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                        @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                                        @Nullable @JsonProperty("bindTestedServerToUnboundNodes") Boolean bindTestedServerToUnboundNodes,
                                        @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                        @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, providedContentSource, bindTestedServerToUnboundNodes,
                    autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.PROVIDE_FILE;
            assert this.getSource() == RequestRSpecSource.PROVIDE_FILE;
        }
    }

    @JsonTypeInfo(
            use = JsonTypeInfo.Id.NAME,
            include = JsonTypeInfo.As.PROPERTY,
            property = "source",
            defaultImpl = ProvidedContentRequestRSpec.class,
            visible = true)
    @JsonSubTypes({
            @JsonSubTypes.Type(value = ProvidedUrlRequestRSpec.class, name = "PROVIDE_URL"),
            @JsonSubTypes.Type(value = ProvidedFileRequestRSpec.class, name = "PROVIDE_FILE"),
            @JsonSubTypes.Type(value = ProvidedContentRequestRSpec.class, name = "PROVIDE_CONTENT")
    })
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static abstract class ProvidedRequestRSpec extends RequestRSpec {
        @Nullable private final String providedContentSource; //url, file or content as specified in source

        private final boolean bindTestedServerToUnboundNodes;

        public ProvidedRequestRSpec() {
            super(RequestRSpecSource.PROVIDE_CONTENT,
                    false,
                    false);
            this.providedContentSource = "";
            //best that bindTestedServerToUnboundNodes is default. For geni tutorial tests, this is very convenient.
            //Also, jFed Gui Logic doesn't work correctly for unbound nodes this anyway.
            this.bindTestedServerToUnboundNodes = true;
        }

        @JsonCreator
        public ProvidedRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                    @Nullable @JsonProperty("providedContentSource") String providedContentSource,
                                    @Nullable @JsonProperty("bindTestedServerToUnboundNodes") Boolean bindTestedServerToUnboundNodes,
                                    @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                    @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.PROVIDE_CONTENT ||
                    source == RequestRSpecSource.PROVIDE_FILE ||
                    source == RequestRSpecSource.PROVIDE_URL;
            assert this.getSource() == RequestRSpecSource.PROVIDE_CONTENT ||
                    this.getSource() == RequestRSpecSource.PROVIDE_FILE ||
                    this.getSource() == RequestRSpecSource.PROVIDE_URL;
            this.providedContentSource = providedContentSource;
            this.bindTestedServerToUnboundNodes = bindTestedServerToUnboundNodes == null ?
                    new ProvidedContentRequestRSpec().getBindTestedServerToUnboundNodes() :
                    bindTestedServerToUnboundNodes;
        }

        @JsonIgnore
        public boolean isValid() {
            if (providedContentSource == null)
                throw new IllegalArgumentException("Config Error: rspec.providedContentSource may not be null if rspec.source == "+source);
            if (providedContentSource.trim().isEmpty())
                throw new IllegalArgumentException("Config Error: rspec.providedContentSource may not be empty if rspec.source == "+source);
            return true;
        }

        @JsonProperty
        @Nullable
        public String getProvidedContentSource() {
            return providedContentSource;
        }

        @JsonProperty
        public boolean getBindTestedServerToUnboundNodes() {
            return bindTestedServerToUnboundNodes;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof ProvidedRequestRSpec)) return false;
            if (!super.equals(o)) return false;

            ProvidedRequestRSpec that = (ProvidedRequestRSpec) o;

            if (bindTestedServerToUnboundNodes != that.bindTestedServerToUnboundNodes) return false;
            return providedContentSource != null ? providedContentSource.equals(that.providedContentSource) : that.providedContentSource == null;
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            result = 31 * result + (providedContentSource != null ? providedContentSource.hashCode() : 0);
            result = 31 * result + (bindTestedServerToUnboundNodes ? 1 : 0);
            return result;
        }

        @Override
        public String toString() {
            return "ProvidedRequestRSpec{" +
                    "source='" + getSource() + '\'' +
                    ", providedContentSource=" + providedContentSource +
                    ", bindTestedServerToUnboundNodes=" + bindTestedServerToUnboundNodes +
                    '}';
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class GenerateUsingAdvertisementRequestRSpec extends RequestRSpec {
        @Nonnull private final AdvertisementBasedRspecGeneratorConfig
                config;

        public GenerateUsingAdvertisementRequestRSpec() {
            super(RequestRSpecSource.GENERATE_USING_ADVERTISEMENT, false, false);
            this.config = new AdvertisementBasedRspecGeneratorConfig(Collections.singletonList(
                    new AdvertisementBasedRspecGeneratorConfig.NodeInfo(
                            1, null,"node", null,
                            null, null, null,
                            null, null, null,
                            null, null, true, null
                    )
            ));
        }

        @JsonCreator
        public GenerateUsingAdvertisementRequestRSpec(
                @Nonnull @JsonProperty("source") RequestRSpecSource source,
                @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease,
                @Nullable @JsonProperty("config") AdvertisementBasedRspecGeneratorConfig config) {
            super(source, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.GENERATE_USING_ADVERTISEMENT;
            assert this.getSource() == RequestRSpecSource.GENERATE_USING_ADVERTISEMENT;
            this.config =
                    config == null ?
                            new GenerateUsingAdvertisementRequestRSpec().getConfig() :
                            config;
        }

        @JsonIgnore
        public boolean isValid() {
            if (config == null)
                throw new IllegalArgumentException("Config Error: rspec.config required for rspec.source == "+source);
            if (config.getNodes().isEmpty())
                throw new IllegalArgumentException("Config Error: rspec.config.nodes may not be empty for rspec.source == "+source);
            return true;
        }

        @JsonProperty
        @Nonnull
        public AdvertisementBasedRspecGeneratorConfig getConfig() {
            return config;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof GenerateUsingAdvertisementRequestRSpec)) return false;
            if (!super.equals(o)) return false;

            GenerateUsingAdvertisementRequestRSpec that = (GenerateUsingAdvertisementRequestRSpec) o;

            //noinspection ConstantConditions
            return config != null ? config.equals(that.config) : that.config == null;
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            //noinspection ConstantConditions
            result = 31 * result + (config != null ? config.hashCode() : 0);
            return result;
        }

        @Override
        public String toString() {
            return "GenerateUsingAdvertisementRequestRSpec{" +
                    "source=" + this.getSource() +
                    ", autoAssignSpecificNodesUsingListResources=" + this.getAutoAssignSpecificNodesUsingListResources() +
                    ", useNitosLease=" + this.getUseNitosLease() +
                    ", config=" + config +
                    '}';
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class GenerateSimpleRequestRSpec extends RequestRSpec {
        private final int nodeCount;
        @Nonnull private final String nodeSliverType;
        private final boolean nodeExclusive;
        @Nonnull private final List<String> nodeComponentUrn;
        @Nonnull private final List<String> nodeAnsibleGroup;

        public GenerateSimpleRequestRSpec() {
            super(RequestRSpecSource.GENERATE_SIMPLE, false, false);
            this.nodeCount = 1;
            this.nodeSliverType = "raw-pc";
            this.nodeExclusive = true;
            this.nodeComponentUrn = Collections.emptyList();
            this.nodeAnsibleGroup = Collections.emptyList();
        }

        @JsonCreator
        public GenerateSimpleRequestRSpec(@Nonnull @JsonProperty("source") RequestRSpecSource source,
                                          @Nullable @JsonProperty("nodeCount") Integer nodeCount,
                                          @Nullable @JsonProperty("nodeSliverType") String nodeSliverType,
                                          @Nullable @JsonProperty("nodeExclusive") Boolean nodeExclusive,
                                          @Nullable @JsonProperty("nodeComponentUrn") List<String> nodeComponentUrn,
                                          @Nullable @JsonProperty("nodeAnsibleGroup") List<String> nodeAnsibleGroup,
                                          @Nullable @JsonProperty("autoAssignSpecificNodesUsingListResources") Boolean autoAssignSpecificNodesUsingListResources,
                                          @Nullable @JsonProperty("useNitosLease") Boolean useNitosLease) {
            super(source, autoAssignSpecificNodesUsingListResources, useNitosLease);
            assert source == RequestRSpecSource.GENERATE_SIMPLE;
            assert this.getSource() == RequestRSpecSource.GENERATE_SIMPLE;

            this.nodeCount = nodeCount == null ? new GenerateSimpleRequestRSpec().getNodeCount() : nodeCount;
            this.nodeSliverType = nodeSliverType == null ? new GenerateSimpleRequestRSpec().getNodeSliverType() : nodeSliverType;
            this.nodeExclusive = nodeExclusive == null ? new GenerateSimpleRequestRSpec().getNodeExclusive() : nodeExclusive;
            this.nodeComponentUrn = nodeComponentUrn == null ? new GenerateSimpleRequestRSpec().getNodeComponentUrn() : nodeComponentUrn;
            this.nodeAnsibleGroup = nodeAnsibleGroup == null ? new GenerateSimpleRequestRSpec().getNodeAnsibleGroup() : nodeAnsibleGroup;
        }

        @JsonIgnore
        public boolean isValid() {
            if (nodeCount <= 0)
                throw new IllegalArgumentException("Config Error: rspec.nodeCount == "+nodeCount+" makes no sense if rspec.source == "+source);
            return true;
        }

        @JsonProperty
        public int getNodeCount() {
            return nodeCount;
        }

        @JsonProperty
        @Nonnull
        public String getNodeSliverType() {
            return nodeSliverType;
        }

        @JsonProperty
        public boolean getNodeExclusive() {
            return nodeExclusive;
        }

        @JsonProperty
        @Nonnull
        public List<String> getNodeComponentUrn() {
            return nodeComponentUrn;
        }

        @JsonProperty
        @Nonnull
        public List<String> getNodeAnsibleGroup() {
            return nodeAnsibleGroup;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof GenerateSimpleRequestRSpec)) return false;
            if (!super.equals(o)) return false;

            GenerateSimpleRequestRSpec that = (GenerateSimpleRequestRSpec) o;

            if (nodeCount != that.nodeCount) return false;
            if (nodeExclusive != that.nodeExclusive) return false;
            if (!nodeSliverType.equals(that.nodeSliverType)) return false;
            if (!nodeAnsibleGroup.equals(that.nodeAnsibleGroup)) return false;
            return nodeComponentUrn.equals(that.nodeComponentUrn);
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            result = 31 * result + nodeCount;
            result = 31 * result + nodeSliverType.hashCode();
            result = 31 * result + (nodeExclusive ? 1 : 0);
            result = 31 * result + nodeComponentUrn.hashCode();
            result = 31 * result + nodeAnsibleGroup.hashCode();
            return result;
        }
    }

    public enum ProjectSource { NONE, AUTO, PROVIDED }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Slice {
        @Nullable private final String sliceName;
        private final boolean failOnExistingSlice;
        private final boolean failOnNonExistingSlice;
        private final long expireTimeMin;
        private final boolean renewExistingSliceIfNeeded;

        @Nonnull private final ProjectSource projectSource;
        private final boolean failIfNoProject; //if projectSource is AUTO or PROVIDED
        @Nullable private final String project; //if projectSource is PROVIDED

        public Slice() {
            this.sliceName = null;
            this.failOnExistingSlice = true;
            this.failOnNonExistingSlice = false;
            this.expireTimeMin = 60;
            this.renewExistingSliceIfNeeded = true;
            this.projectSource = ProjectSource.AUTO;
            this.failIfNoProject = false;
            this.project = null;
        }

        @JsonCreator
        public Slice(@Nullable @JsonProperty("sliceName") String sliceName,
                     @Nullable @JsonProperty("failOnExistingSlice") Boolean failOnExistingSlice,
                     @Nullable @JsonProperty("failOnNonExistingSlice") Boolean failOnNonExistingSlice,
                     @Nullable @JsonProperty("expireTimeMin") Long expireTimeMin,
                     @Nullable @JsonProperty("renewExistingSliceIfNeeded") Boolean renewExistingSliceIfNeeded,
                     @Nullable @JsonProperty("projectSource") ProjectSource projectSource,
                     @Nullable @JsonProperty("failIfNoProject") Boolean failIfNoProject,
                     @Nullable @JsonProperty("project") String project) {
            this.sliceName = sliceName == null ? new Slice().getSliceName() : sliceName;
            this.failOnExistingSlice = failOnExistingSlice == null ? new Slice().getFailOnExistingSlice() : failOnExistingSlice;
            this.failOnNonExistingSlice = failOnNonExistingSlice == null ? new Slice().getFailOnNonExistingSlice() : failOnNonExistingSlice;
            this.expireTimeMin = expireTimeMin == null ? new Slice().getExpireTimeMin() : expireTimeMin;
            this.renewExistingSliceIfNeeded = renewExistingSliceIfNeeded == null ? new Slice().getRenewExistingSliceIfNeeded() : renewExistingSliceIfNeeded;
            this.projectSource = projectSource == null ? new Slice().getProjectSource() : projectSource;
            this.failIfNoProject = failIfNoProject == null ? new Slice().getFailIfNoProject() : failIfNoProject;
            this.project = project == null ? new Slice().getProject() : project;
        }

        @JsonIgnore
        public boolean isValid() {
            if (expireTimeMin <= 0)
                throw new IllegalArgumentException("Config Error: slice.expireTimeMin <= makes no sense: "+expireTimeMin);
            if (projectSource == ProjectSource.PROVIDED && project == null && failIfNoProject)
                throw new IllegalArgumentException("Config Error: Combination of failIfNoProject=true and project==null and projectSource == ProjectSource.PROVIDED will always fail");
            if (projectSource == ProjectSource.NONE && failIfNoProject)
                throw new IllegalArgumentException("Config Error: Combination of failIfNoProject=true and projectSource == ProjectSource.NONE will always fail");
            if (failOnExistingSlice && failOnNonExistingSlice)
                throw new IllegalArgumentException("Config Error: failOnExistingSlice and failOnNonExistingSlice should not both be true");

            return true;
        }

        @Nullable
        public String getSliceName() {
            return sliceName;
        }

        public boolean getFailOnExistingSlice() {
            return failOnExistingSlice;
        }

        public boolean getFailOnNonExistingSlice() {
            return failOnNonExistingSlice;
        }

        public long getExpireTimeMin() {
            return expireTimeMin;
        }

        public boolean getRenewExistingSliceIfNeeded() {
            return renewExistingSliceIfNeeded;
        }

        @Nonnull
        public ProjectSource getProjectSource() {
            return projectSource;
        }

        public boolean getFailIfNoProject() {
            return failIfNoProject;
        }

        @Nullable
        public String getProject() {
            return project;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Slice)) return false;

            Slice slice = (Slice) o;

            if (failOnExistingSlice != slice.failOnExistingSlice) return false;
            if (failOnNonExistingSlice != slice.failOnNonExistingSlice) return false;
            if (expireTimeMin != slice.expireTimeMin) return false;
            if (renewExistingSliceIfNeeded != slice.renewExistingSliceIfNeeded) return false;
            if (failIfNoProject != slice.failIfNoProject) return false;
            if (sliceName != null ? !sliceName.equals(slice.sliceName) : slice.sliceName != null) return false;
            if (projectSource != slice.projectSource) return false;
            return project != null ? project.equals(slice.project) : slice.project == null;
        }

        @Override
        public int hashCode() {
            int result = sliceName != null ? sliceName.hashCode() : 0;
            result = 31 * result + (failOnExistingSlice ? 1 : 0);
            result = 31 * result + (failOnNonExistingSlice ? 1 : 0);
            result = 31 * result + (int) (expireTimeMin ^ (expireTimeMin >>> 32));
            result = 31 * result + (renewExistingSliceIfNeeded ? 1 : 0);
            result = 31 * result + projectSource.hashCode();
            result = 31 * result + (failIfNoProject ? 1 : 0);
            result = 31 * result + (project != null ? project.hashCode() : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Provision {
        //Allocate, Provision, POA    (skipped if requestRspec method is "already_povisioned")
//        private final long provisionExpireTimeMin;
        private final boolean useGeniEndTimeOptionForCreateSliver;
        private final boolean enabled;

        public Provision() {
//            this.provisionExpireTimeMin = 25;
            this.useGeniEndTimeOptionForCreateSliver = false;
            this.enabled = true;
        }
        @JsonCreator
        public Provision(
//                @Nullable @JsonProperty("provisionExpireTimeMin") Long provisionExpireTimeMin,
                @Nullable @JsonProperty("useGeniEndTimeOptionForCreateSliver") Boolean useGeniEndTimeOptionForCreateSliver,
                @Nullable @JsonProperty("enabled") Boolean enabled
        ) {
//            this.provisionExpireTimeMin = provisionExpireTimeMin == null ? new Provision().getProvisionExpireTimeMin() : provisionExpireTimeMin;
            this.useGeniEndTimeOptionForCreateSliver = useGeniEndTimeOptionForCreateSliver == null ? new Provision().getUseGeniEndTimeOptionForCreateSliver() : useGeniEndTimeOptionForCreateSliver;
            this.enabled = enabled == null ? new Provision().isEnabled() : enabled;
        }

        @JsonIgnore
        public boolean isValid() {
            return true;
        }

//        @JsonProperty
//        public long getProvisionExpireTimeMin() {
//            return provisionExpireTimeMin;
//        }

        @JsonProperty
        public boolean getUseGeniEndTimeOptionForCreateSliver() {
            return useGeniEndTimeOptionForCreateSliver;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Provision)) return false;

            Provision provision = (Provision) o;

            if (useGeniEndTimeOptionForCreateSliver != provision.useGeniEndTimeOptionForCreateSliver) return false;
            return enabled == provision.enabled;
        }

        @Override
        public int hashCode() {
            int result = (useGeniEndTimeOptionForCreateSliver ? 1 : 0);
            result = 31 * result + (enabled ? 1 : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class WaitForReady {
        private final long maxTimeMin;
        private final boolean skipFinalManifestCheck;
        private final boolean requireNodesInManifest;
        private final long extraWaitTimeSeconds;
        private final boolean enabled;

        public WaitForReady() {
            this.maxTimeMin = 20;
            this.skipFinalManifestCheck = false;
            this.requireNodesInManifest = true;
            this.extraWaitTimeSeconds = 0;
            this.enabled = true;
        }
        @JsonCreator
        public WaitForReady(@Nullable @JsonProperty("maxTimeMin") Long maxTimeMin,
                            @Nullable @JsonProperty("skipFinalManifestCheck") Boolean skipFinalManifestCheck,
                            @Nullable @JsonProperty("requireNodesInManifest") Boolean requireNodesInManifest,
                            @Nullable @JsonProperty("extraWaitTimeSeconds") Long extraWaitTimeSeconds,
                            @Nullable @JsonProperty("enabled") Boolean enabled) {
            this.maxTimeMin = maxTimeMin == null ? new WaitForReady().getMaxTimeMin() : maxTimeMin;
            this.skipFinalManifestCheck = skipFinalManifestCheck == null ? new WaitForReady().getSkipFinalManifestCheck() : skipFinalManifestCheck;
            this.requireNodesInManifest = requireNodesInManifest == null ? new WaitForReady().getRequireNodesInManifest() : requireNodesInManifest;
            this.extraWaitTimeSeconds = extraWaitTimeSeconds == null ? new WaitForReady().getExtraWaitTimeSeconds() : extraWaitTimeSeconds;
            this.enabled = enabled == null ? new WaitForReady().isEnabled() : enabled;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!enabled)
                return true;

            if (maxTimeMin < 0)
                throw new IllegalArgumentException("Config Error: waitForReady.maxTimeMin < 0 makes no sense");

            return true;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Skip the detailed manifest test in the \"checkManifest\" step? If skipped, very basic testing of the manifest is still done.")
        public boolean getSkipFinalManifestCheck() {
            return skipFinalManifestCheck;
        }


        @JsonProperty
        public long getMaxTimeMin() {
            return maxTimeMin;
        }

        @JsonProperty
        public boolean getRequireNodesInManifest() {
            return requireNodesInManifest;
        }

        @JsonProperty
        public long getExtraWaitTimeSeconds() {
            return extraWaitTimeSeconds;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof WaitForReady)) return false;

            WaitForReady that = (WaitForReady) o;

            if (maxTimeMin != that.maxTimeMin) return false;
            if (skipFinalManifestCheck != that.skipFinalManifestCheck) return false;
            if (requireNodesInManifest != that.requireNodesInManifest) return false;
            if (extraWaitTimeSeconds != that.extraWaitTimeSeconds) return false;
            return enabled == that.enabled;
        }

        @Override
        public int hashCode() {
            int result = (int) (maxTimeMin ^ (maxTimeMin >>> 32));
            result = 31 * result + (skipFinalManifestCheck ? 1 : 0);
            result = 31 * result + (requireNodesInManifest ? 1 : 0);
            result = 31 * result + (int) (extraWaitTimeSeconds ^ (extraWaitTimeSeconds >>> 32));
            result = 31 * result + (enabled ? 1 : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Cleanup {
        private final boolean enabled;
        private final boolean ignoreErrors;
        private final long maxWaitTimeMinutes;

        public Cleanup() {
            this.enabled = true;
            this.ignoreErrors = false;
            this.maxWaitTimeMinutes = 5;
        }

        @JsonCreator
        public Cleanup(@Nullable @JsonProperty("enabled") Boolean enabled,
                       @Nullable @JsonProperty("ignoreErrors") Boolean ignoreErrors,
                       @Nullable @JsonProperty("maxWaitTimeMinutes") Long maxWaitTimeMinutes) {
            this.enabled = enabled == null ? new Cleanup().isEnabled() : enabled;
            this.ignoreErrors = ignoreErrors == null ? new Cleanup().getIgnoreErrors() : ignoreErrors;
            this.maxWaitTimeMinutes = maxWaitTimeMinutes == null ? new Cleanup().getMaxWaitTimeMinutes() : maxWaitTimeMinutes;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!enabled)
                return true;

            if (maxWaitTimeMinutes < 0)
                throw new IllegalArgumentException("Config Error: cleanup.maxWaitTimeMinutes < 0 makes no sense");

            return true;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @JsonProperty
        public boolean getIgnoreErrors() {
            return ignoreErrors;
        }

        @JsonProperty
        public long getMaxWaitTimeMinutes() {
            return maxWaitTimeMinutes;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Cleanup)) return false;

            Cleanup cleanup = (Cleanup) o;

            if (enabled != cleanup.enabled) return false;
            if (ignoreErrors != cleanup.ignoreErrors) return false;
            return maxWaitTimeMinutes == cleanup.maxWaitTimeMinutes;
        }

        @Override
        public int hashCode() {
            int result = (enabled ? 1 : 0);
            result = 31 * result + (ignoreErrors ? 1 : 0);
            result = 31 * result + (int) (maxWaitTimeMinutes ^ (maxWaitTimeMinutes >>> 32));
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Resource {
        @Nonnull private final Slice slice;
        @Nullable private final RequestRSpec requestRSpec;
        @Nullable private final ESpecTestConfig.ESpec eSpec;
        @Nonnull private final Provision provision;
        @Nonnull private final WaitForReady waitForReady;
        @Nonnull private final Cleanup cleanup;
        private final boolean overrideESpecRSpec;

        public Resource() {
            this.slice = new Slice();
            this.requestRSpec = RequestRSpec.getDefault();
            this.eSpec = null;
            this.provision = new Provision();
            this.waitForReady = new WaitForReady();
            this.cleanup = new Cleanup();
            this.overrideESpecRSpec = false;
        }
        @JsonCreator
        public Resource(@Nullable @JsonProperty("slice") Slice slice,
                        @Nullable @JsonProperty("requestRSpec") RequestRSpec requestRSpec,
                        @Nullable @JsonProperty("eSpec") ESpecTestConfig.ESpec eSpec,
                        @Nullable @JsonProperty("provision") Provision provision,
                        @Nullable @JsonProperty("waitForReady") WaitForReady waitForReady,
                        @Nullable @JsonProperty("cleanup") Cleanup cleanup,
                        @Nullable @JsonProperty("overrideESpecRSpec") Boolean overrideESpecRSpec) {
            this.slice = slice == null ? new Resource().getSlice() : slice;
            this.requestRSpec = requestRSpec == null ?
                    (eSpec == null ? new Resource().getRequestRSpec() : null) :
                    requestRSpec;
            this.eSpec = eSpec == null ? null : eSpec;
            this.provision = provision == null ? new Resource().getProvision() : provision;
            this.waitForReady = waitForReady == null ? new Resource().getWaitForReady() : waitForReady;
            this.cleanup = cleanup == null ? new Resource().getCleanup() : cleanup;
            this.overrideESpecRSpec = overrideESpecRSpec == null ? new Resource().getOverrideESpecRSpec() : overrideESpecRSpec;

            if (requestRSpec != null && eSpec != null) {
                throw new IllegalArgumentException("requestRSpec and eSpec should never both be specified");
            }
        }

        @JsonIgnore
        public boolean isValid() {
            boolean res = getSlice().isValid();
            RequestRSpec requestRspec = getRequestRSpec();
            ESpecTestConfig.ESpec espec = getESpec();
            res = res && (requestRspec == null || requestRspec.isValid());
            res = res && (espec == null || espec.isValid());
            res = res && (espec != null || requestRspec != null);
            res = res && getProvision().isValid();
            res = res && getWaitForReady().isValid();
            res = res && getCleanup().isValid();
            if (!getProvision().isEnabled() && !getWaitForReady().isEnabled())
                throw new IllegalArgumentException("Config Error: if the Provision step is disabled, waitForReady should be disabled too.");
            if (getOverrideESpecRSpec()) {
                if (espec == null || requestRspec == null)
                    throw new IllegalArgumentException("Config Error: overrideESpecRSpec=true requires specifying both an RSpec and an ESpec");
            } else {
                if (espec != null && requestRspec != null)
                    throw new IllegalArgumentException("Config Error: Specifying both an RSpec and an ESpec requires overrideESpecRSpec=true");
            }
            return res;
        }

        @JsonProperty
        @Nonnull
        public Slice getSlice() {
            return slice;
        }

        @JsonProperty
        @Nullable
        @ApiTestConfigProperty(noEdit = true) //do not edit -> type is too complex
        public RequestRSpec getRequestRSpec() {
            return requestRSpec;
        }

        @JsonProperty("eSpec")
        @Nullable
        public ESpecTestConfig.ESpec getESpec() {
            return eSpec;
        }

        @JsonProperty
        @Nonnull
        public Provision getProvision() {
            return provision;
        }

        @JsonProperty
        @Nonnull
        public WaitForReady getWaitForReady() {
            return waitForReady;
        }

        @JsonProperty
        @Nonnull
        public Cleanup getCleanup() {
            return cleanup;
        }

        @JsonProperty("overrideESpecRSpec")
        public boolean getOverrideESpecRSpec() {
            return overrideESpecRSpec;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Resource)) return false;
            Resource resource = (Resource) o;
            return Objects.equals(slice, resource.slice) &&
                    Objects.equals(requestRSpec, resource.requestRSpec) &&
                    Objects.equals(eSpec, resource.eSpec) &&
                    Objects.equals(provision, resource.provision) &&
                    Objects.equals(waitForReady, resource.waitForReady) &&
                    Objects.equals(cleanup, resource.cleanup) &&
                    overrideESpecRSpec == resource.overrideESpecRSpec;
        }

        @Override
        public int hashCode() {
            return Objects.hash(slice, requestRSpec, eSpec, provision, waitForReady, cleanup, overrideESpecRSpec);
        }
    }




    public NodeLoginTestStepConfig generateNodeLoginTestStepConfig(@Nonnull GeniUserProvider geniUserProvider) {
        NodeLoginTestStepConfig.SSHKeySource sshKeySource;
        PublicKey publicKey;
        PrivateKey privateKey;
        switch (getSshKeys().getSource()) {
            case RANDOM: {
                sshKeySource = NodeLoginTestStepConfig.SSHKeySource.RANDOM;
                publicKey = null;
                privateKey = null;
                break;
            }
            case TEST_USER: {
                sshKeySource = NodeLoginTestStepConfig.SSHKeySource.PROVIDED;
                assert geniUserProvider.isUserLoggedIn();
                GeniUser user = geniUserProvider.getLoggedInGeniUser();
                assert user != null;
                publicKey = user.getPublicKey();
                privateKey = user.getPrivateKey();
                break;
            }
            default: throw new IllegalArgumentException("Unsupported ssh keys source: "+getSshKeys().getSource());
        }


        return new NodeLoginTestStepConfig(sshKeySource,
                publicKey,
                privateKey,
                getNodeLoginTest().getUseExternalSsh(),
                getNodeLoginTest().getProxy(),
                getNodeLoginTest().getDeadlineSeconds() * 1000l,
                getNodeLoginTest().getPreferredUser(),
                getNodeLoginTest().getForceIpVersion());
    }


    @Contract("null -> null")
    @Nullable private static String trimAndToNullIfEmpty(@Nullable String orig) {
        if (orig == null)
            return null;
        if (orig.trim().isEmpty())
            return null;
        return orig.trim();
    }

    @Contract("null -> null")
    @Nullable private static String toNullIfEmpty(@Nullable String orig) {
        if (orig == null)
            return null;
        if (orig.trim().isEmpty())
            return null;
        return orig.trim();
    }

    @FunctionalInterface
    public interface NonNullSupplier<T> {
        @Nonnull T get();
    }

    @Contract("null,null -> null; _,!null->!null")
    @Nullable private static String trimAndToDefaultIfEmpty(@Nullable String orig,
                                                            @Nullable NonNullSupplier<String> defaultString) {
        if (orig == null)
            return defaultString == null ? null : defaultString.get();
        if (orig.trim().isEmpty())
            return defaultString == null ? null : defaultString.get();
        return orig.trim();
    }

    @Contract("null,null -> null; _,!null->!null")
    @Nullable private static String toDefaultIfEmpty(@Nullable String orig,
                                                     @Nullable NonNullSupplier<String> defaultString) {
        if (orig == null)
            return defaultString == null ? null : defaultString.get();
        if (orig.trim().isEmpty())
            return defaultString == null ? null : defaultString.get();
        return orig;
    }

    @Contract("null,null -> null")
    @Nullable private static String toNullableDefaultIfEmpty(@Nullable String orig,
                                                             @Nullable Supplier<String> defaultString) {
        if (orig == null)
            return defaultString == null ? null : defaultString.get();
        if (orig.trim().isEmpty())
            return defaultString == null ? null : defaultString.get();
        return orig;
    }
}
